/* (C) 2009-2011,2016 Yana */

#include "yk0b.h"
#include "yk0b_base.h"
#include "yk0b_internal.h"
#include "yk0b_setup.h"
#include <stddef.h>

#if YK0B_SETUP_EXCEPTION_PORT == YK0B_SETUP_INTERRUPT_PORT
#error YK0B_SETUP_EXCEPTION_PORT == YK0B_SETUP_INTERRUPT_PORT
#endif

#define QUQUE_NULL (~0)

/* queue */

typedef struct yk0b_queue {
	yk0b_task_t next;
	yk0b_task_t previous;
} yk0b_queue_t;

/* type */

typedef yk0b_uint32_t yk0b_state_t;

typedef struct yk0b_task_control {
	yk0b_queue_t queue;
	yk0b_uint32_t context[YK0B_CONTEXT_SIZE];
	yk0b_priority_t priority;
	yk0b_state_t state;
	union {
		yk0b_port_t port;
		yk0b_task_t replyer;
	} u;
} yk0b_task_control_t;

typedef yk0b_task_t yk0b_task_link_t;

typedef struct {
	yk0b_queue_t queue;
	yk0b_flag_t flag;
	yk0b_task_link_t receiver;
} yk0b_port_control_t;

/* variable */

yk0b_port_control_t yk0b_port_control[YK0B_SETUP_PORT_SIZE];
yk0b_task_link_t yk0b_schedule_link;
yk0b_queue_t yk0b_schedule_queue[YK0B_SETUP_PRIORITY_SIZE];
yk0b_task_control_t yk0b_task_control[YK0B_SETUP_TASK_SIZE];

/* queue */

static void yk0b_queue_append(yk0b_queue_t *root, yk0b_task_t node) {
	yk0b_task_control[node].queue.next = QUQUE_NULL;
	yk0b_task_control[node].queue.previous = root->previous;
	if (root->previous == QUQUE_NULL) {
		root->next = node;
	} else {
		yk0b_task_control[root->previous].queue.next = node;
	}
	root->previous = node;
}

yk0b_task_t yk0b_queue_copy(yk0b_queue_t *root) {
	return root->next;
}

static void yk0b_queue_delete(yk0b_queue_t *root, yk0b_task_t node) {
	if (yk0b_task_control[node].queue.next == QUQUE_NULL) {
		root->previous = yk0b_task_control[node].queue.previous;
	} else {
		yk0b_task_control[yk0b_task_control[node].queue.next].queue.previous = yk0b_task_control[node].queue.previous;
	}
	if (yk0b_task_control[node].queue.previous == QUQUE_NULL) {
		root->next = yk0b_task_control[node].queue.next;
	} else {
		yk0b_task_control[yk0b_task_control[node].queue.previous].queue.next = yk0b_task_control[node].queue.next;
	}
}

static void yk0b_queue_initialize(yk0b_queue_t *root) {
	root->next = QUQUE_NULL;
	root->previous = QUQUE_NULL;
}

static yk0b_task_t yk0b_queue_remove(yk0b_queue_t *root) {
	if (root->next == QUQUE_NULL) {
		return QUQUE_NULL;
	} else {
		yk0b_task_t node;
		node = root->next;
		root->next = yk0b_task_control[node].queue.next;
		if (yk0b_task_control[node].queue.next == QUQUE_NULL) {
			root->previous = yk0b_task_control[node].queue.previous;
		} else {
			yk0b_task_control[yk0b_task_control[node].queue.next].queue.previous = yk0b_task_control[node].queue.previous;
		}
		return node;
	}
}

/* access function */

static void yk0b_port_append(yk0b_port_t port, yk0b_task_link_t sender) {
	yk0b_queue_append((yk0b_queue_t *)&yk0b_port_control[port], sender);
	yk0b_port_control[port].flag |= YK0B_FLAG_KERNEL_SENDER;
}

static void yk0b_port_delete(yk0b_port_t port, yk0b_task_link_t sender) {
	yk0b_queue_delete((yk0b_queue_t *)&yk0b_port_control[port], sender);
	if (yk0b_queue_copy((yk0b_queue_t *)&yk0b_port_control[port]) == QUQUE_NULL) {
		yk0b_port_control[port].flag &= ~YK0B_FLAG_KERNEL_SENDER;
	}
}

static void yk0b_port_initialize(void) {
	yk0b_port_t port;
	for (port = 0; port < YK0B_SETUP_PORT_SIZE; ++port) {
		yk0b_queue_initialize((yk0b_queue_t *)&yk0b_port_control[port]);
		yk0b_port_control[port].flag = 0;
	}
}

static yk0b_flag_t yk0b_port_read_flag(yk0b_port_t port) {
	return yk0b_port_control[port].flag;
}

static yk0b_task_link_t yk0b_port_read_receiver(yk0b_port_t port) {
	return yk0b_port_control[port].receiver;
}

static yk0b_task_link_t yk0b_port_remove(yk0b_port_t port) {
	yk0b_task_link_t sender;
	sender = (yk0b_task_link_t)yk0b_queue_remove((yk0b_queue_t *)&yk0b_port_control[port]);
	if (yk0b_queue_copy((yk0b_queue_t *)&yk0b_port_control[port]) == QUQUE_NULL) {
		yk0b_port_control[port].flag &= ~YK0B_FLAG_KERNEL_SENDER;
	}
	return sender;
}

static void yk0b_port_write_flag(yk0b_port_t port, yk0b_flag_t flag) {
	yk0b_port_control[port].flag = flag;
}

static void yk0b_port_write_receiver(yk0b_port_t port, yk0b_task_link_t receiver) {
	yk0b_port_control[port].receiver = receiver;
}

static void yk0b_schedule_append(yk0b_task_link_t link) {
	yk0b_queue_append(&yk0b_schedule_queue[yk0b_task_control[link].priority], link);
}

static void yk0b_schedule_delete(yk0b_task_link_t link) {
	yk0b_queue_delete(&yk0b_schedule_queue[yk0b_task_control[link].priority], link);
}

static void yk0b_schedule_initialize(void) {
	yk0b_priority_t priority;
	yk0b_schedule_link = QUQUE_NULL;
	for (priority = 0; priority < YK0B_SETUP_PRIORITY_SIZE; ++priority) {
		yk0b_queue_initialize(&yk0b_schedule_queue[priority]);
	}
}

static void yk0b_schedule_update(void) {
	yk0b_priority_t priority;
	for (priority = 0; priority < YK0B_SETUP_PRIORITY_SIZE; ++priority) {
		yk0b_schedule_link = (yk0b_task_link_t)yk0b_queue_copy(&yk0b_schedule_queue[priority]);
		if (yk0b_schedule_link != QUQUE_NULL) {
			return;
		}
	}
}

static yk0b_task_t yk0b_task_link_to_task(yk0b_task_link_t link) {
	return link;
}

static yk0b_port_t yk0b_task_read_port(yk0b_task_link_t link) {
	return yk0b_task_control[link].u.port;
}

static yk0b_priority_t yk0b_task_read_priority(yk0b_task_link_t link) {
	return yk0b_task_control[link].priority;
}

static yk0b_task_link_t yk0b_task_read_replyer(yk0b_task_link_t link) {
	return yk0b_task_control[link].u.replyer;
}

static yk0b_state_t yk0b_task_read_state(yk0b_task_link_t link) {
	return yk0b_task_control[link].state;
}

static yk0b_task_link_t yk0b_task_to_task_link(yk0b_task_t task) {
	return task;
}

static void yk0b_task_write_port(yk0b_task_link_t link, yk0b_port_t port) {
	yk0b_task_control[link].u.port = port;
}

static void yk0b_task_write_priority(yk0b_task_link_t link, yk0b_priority_t priority) {
	yk0b_task_control[link].priority = priority;
}

static void yk0b_task_write_replyer(yk0b_task_link_t link, yk0b_task_link_t replyer) {
	yk0b_task_control[link].u.replyer = replyer;
}

static void yk0b_task_write_state(yk0b_task_link_t link, yk0b_state_t state) {
	yk0b_task_control[link].state = state;
}

/* context function */

static yk0b_context_value_t yk0b_task_read_context(yk0b_task_link_t link, yk0b_context_type_t type) {
	return yk0b_task_control[link].context[type];
}

static void yk0b_task_write_context(yk0b_task_link_t link, yk0b_context_type_t type, yk0b_context_value_t value) {
	yk0b_task_control[link].context[type] = value;
}

static yk0b_port_t yk0b_accept_read_port(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static void yk0b_accept_write_task(yk0b_task_link_t link, yk0b_task_t task) {
	yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1] = task;
}

static yk0b_port_t yk0b_bit_flag_read_port(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_flag_t yk0b_bit_flag_read_flag_off(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT2];
}

static yk0b_flag_t yk0b_bit_flag_read_flag_on(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT3];
}

static void yk0b_bit_flag_write_flag(yk0b_task_link_t link, yk0b_flag_t flag) {
	yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1] = flag;
}

static yk0b_port_t yk0b_call_read_port(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_port_t yk0b_connect_read_task_replyee(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_task_t yk0b_connect_read_task_replyer(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT2];
}

static yk0b_port_t yk0b_forward_read_port(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_task_t yk0b_forward_read_task(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT2];
}

static yk0b_task_t yk0b_get_context_read_task(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_context_type_t yk0b_get_context_read_type(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT2];
}

static void yk0b_get_context_write_value(yk0b_task_link_t link, yk0b_context_value_t value) {
	yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1] = value;
}

static yk0b_task_t yk0b_get_priority_read_task(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static void yk0b_get_priority_write_priority(yk0b_task_link_t link, yk0b_priority_t priority) {
	yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1] = priority;
}

static yk0b_port_t yk0b_listen_read_port(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_flag_t yk0b_listen_read_flag(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT2];
}

static void yk0b_listen_write_flag(yk0b_task_link_t link, yk0b_flag_t flag) {
	yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1] = flag;
}

static yk0b_task_t yk0b_reply_read_task(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_task_t yk0b_set_context_read_task(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_context_type_t yk0b_set_context_read_type(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT2];
}

static yk0b_context_value_t yk0b_set_context_read_value(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT3];
}

static void yk0b_set_context_write_value(yk0b_task_link_t link, yk0b_context_value_t value) {
	yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1] = value;
}

static yk0b_task_t yk0b_set_priority_read_task(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1];
}

static yk0b_priority_t yk0b_set_priority_read_priority(yk0b_task_link_t link) {
	return yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT2];
}

static void yk0b_set_priority_write_priority(yk0b_task_link_t link, yk0b_priority_t priority) {
	yk0b_task_control[link].context[YK0B_CONTEXT_ARGUMENT1] = priority;
}

/* service function */

static void yk0b_exception_forward(yk0b_port_t port, yk0b_task_link_t link) {
	yk0b_schedule_delete(link);
	yk0b_task_write_port(link, port);
	yk0b_task_write_state(link, YK0B_STATE_RECEIVE);
	yk0b_port_append(port, link);
	if ((yk0b_port_read_flag(port) & YK0B_FLAG_KERNEL_RECEIVER) != 0 && (yk0b_listen_read_flag(yk0b_port_read_receiver(port)) & YK0B_FLAG_KERNEL_SENDER) != 0) {
		yk0b_flag_t f;
		f = yk0b_listen_read_flag(yk0b_port_read_receiver(port));
		yk0b_listen_write_flag(yk0b_port_read_receiver(port), yk0b_port_read_flag(port));
		yk0b_port_write_flag(port, yk0b_port_read_flag(port) & (~(f | YK0B_FLAG_KERNEL_RECEIVER) | YK0B_FLAG_KERNEL_SENDER));
		yk0b_task_write_state(yk0b_port_read_receiver(port), YK0B_STATE_READY);
		yk0b_schedule_append(yk0b_port_read_receiver(port));
	}
	yk0b_schedule_update();
}

static void yk0b_interrupt_bit_flag(yk0b_port_t port, yk0b_flag_t flag_off, yk0b_flag_t flag_on) {
	if ((yk0b_port_read_flag(port) & YK0B_FLAG_KERNEL_RECEIVER) == 0 || (yk0b_listen_read_flag(yk0b_port_read_receiver(port)) & flag_on) == 0) {
		yk0b_port_write_flag(port, (yk0b_port_read_flag(port) & ~flag_off) | flag_on);
	} else {
		yk0b_flag_t f;
		f = yk0b_listen_read_flag(yk0b_port_read_receiver(port));
		yk0b_listen_write_flag(yk0b_port_read_receiver(port), (yk0b_port_read_flag(port) & ~flag_off) | flag_on);
		yk0b_port_write_flag(port, ((yk0b_port_read_flag(port) & ~flag_off) | flag_on) & (~(f | YK0B_FLAG_KERNEL_RECEIVER) | YK0B_FLAG_KERNEL_SENDER));
		yk0b_task_write_state(yk0b_port_read_receiver(port), YK0B_STATE_READY);
		yk0b_schedule_append(yk0b_port_read_receiver(port));
		yk0b_schedule_update();
	}
}

static yk0b_error_t yk0b_procedure_accept(yk0b_task_link_t link) {
	yk0b_task_link_t l;
	yk0b_port_t p;
	p = yk0b_accept_read_port(link);
	if (p >= YK0B_SETUP_PORT_SIZE) {
		return 1;
	}
	if ((yk0b_port_read_flag(p) & YK0B_FLAG_KERNEL_SENDER) == 0) {
		return 1;
	}
	l = yk0b_port_remove(p);
	yk0b_task_write_replyer(l, link);
	yk0b_task_write_state(l, YK0B_STATE_REPLY);
	yk0b_accept_write_task(link, yk0b_task_link_to_task(l));
	return 0;
}

static yk0b_error_t yk0b_procedure_bit_flag(yk0b_task_link_t link) {
	yk0b_flag_t off;
	yk0b_flag_t on;
	yk0b_port_t p;
	p = yk0b_bit_flag_read_port(link);
	if (p >= YK0B_SETUP_PORT_SIZE) {
		return 1;
	}
	off = yk0b_bit_flag_read_flag_off(link);
	on = yk0b_bit_flag_read_flag_on(link);
	yk0b_bit_flag_write_flag(link, yk0b_port_read_flag(p));
	yk0b_interrupt_bit_flag(p, off & ~YK0B_FLAG_KERNEL_MASK, on & ~YK0B_FLAG_KERNEL_MASK);
	return 0;
}

static yk0b_error_t yk0b_procedure_call(yk0b_task_link_t link) {
	yk0b_port_t p;
	p = yk0b_call_read_port(link);
	if (p >= YK0B_SETUP_PORT_SIZE) {
		return 1;
	}
	yk0b_exception_forward(p, link);
	return 0;
}

static yk0b_error_t yk0b_procedure_connect(yk0b_task_link_t link) {
	yk0b_task_t replyee;
	yk0b_task_t replyer;
	replyee = yk0b_connect_read_task_replyee(link);
	if (replyee >= YK0B_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0b_task_read_state(yk0b_task_to_task_link(replyee)) != YK0B_STATE_REPLY) {
		return 1;
	}
	if (yk0b_task_read_replyer(yk0b_task_to_task_link(replyee)) != link) {
		return 1;
	}
	replyer = yk0b_connect_read_task_replyer(link);
	if (replyer >= YK0B_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0b_task_read_state(yk0b_task_to_task_link(replyer)) != YK0B_STATE_REPLY) {
		return 1;
	}
	if (yk0b_task_read_replyer(yk0b_task_to_task_link(replyer)) != link) {
		return 1;
	}
	yk0b_task_write_replyer(yk0b_task_to_task_link(replyee), yk0b_task_to_task_link(replyer));
	return 0;
}

static yk0b_error_t yk0b_procedure_forward(yk0b_task_link_t link) {
	yk0b_port_t p;
	yk0b_task_t t;
	p = yk0b_forward_read_port(link);
	if (p >= YK0B_SETUP_PORT_SIZE) {
		return 1;
	}
	t = yk0b_forward_read_task(link);
	if (t >= YK0B_SETUP_TASK_SIZE) {
		return 1;
	}
	switch (yk0b_task_read_state(yk0b_task_to_task_link(t))) {
	case YK0B_STATE_READY:
		if (yk0b_task_to_task_link(t) == link) {
			return 1;
		}
		yk0b_schedule_delete(yk0b_task_to_task_link(t));
		break;
	case YK0B_STATE_RECEIVE:
		yk0b_port_delete(yk0b_task_read_port(yk0b_task_to_task_link(t)), yk0b_task_to_task_link(t));
		break;
	case YK0B_STATE_REPLY:
		if (yk0b_task_read_replyer(yk0b_task_to_task_link(t)) != link) {
			return 1;
		}
		break;
	case YK0B_STATE_SEND:
		yk0b_listen_write_flag(yk0b_task_to_task_link(t), 0);
		yk0b_port_write_flag(yk0b_task_read_port(yk0b_task_to_task_link(t)), yk0b_port_read_flag(yk0b_task_read_port(yk0b_task_to_task_link(t))) & ~YK0B_FLAG_KERNEL_RECEIVER);
		break;
	}
	yk0b_task_write_port(yk0b_task_to_task_link(t), p);
	yk0b_task_write_state(yk0b_task_to_task_link(t), YK0B_STATE_RECEIVE);
	yk0b_port_append(p, yk0b_task_to_task_link(t));
	if ((yk0b_port_read_flag(p) & YK0B_FLAG_KERNEL_RECEIVER) != 0 && (yk0b_listen_read_flag(yk0b_port_read_receiver(p)) & YK0B_FLAG_KERNEL_SENDER) != 0) {
		yk0b_flag_t f;
		f = yk0b_listen_read_flag(yk0b_port_read_receiver(p));
		yk0b_listen_write_flag(yk0b_port_read_receiver(p), yk0b_port_read_flag(p));
		yk0b_port_write_flag(p, yk0b_port_read_flag(p) & (~(f | YK0B_FLAG_KERNEL_RECEIVER) | YK0B_FLAG_KERNEL_SENDER));
		yk0b_task_write_state(yk0b_port_read_receiver(p), YK0B_STATE_READY);
		yk0b_schedule_append(yk0b_port_read_receiver(p));
		yk0b_schedule_update();
	}
	return 0;
}

static yk0b_error_t yk0b_procedure_get_context(yk0b_task_link_t link) {
	yk0b_context_type_t ct;
	yk0b_task_t t;
	t = yk0b_get_context_read_task(link);
	if (t >= YK0B_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0b_task_read_state(yk0b_task_to_task_link(t)) != YK0B_STATE_REPLY) {
		return 1;
	}
	if (yk0b_task_read_replyer(yk0b_task_to_task_link(t)) != link) {
		return 1;
	}
	ct = yk0b_get_context_read_type(link);
	if (ct >= YK0B_CONTEXT_SIZE) {
		return 1;
	}
	yk0b_get_context_write_value(link, yk0b_task_read_context(yk0b_task_to_task_link(t), ct));
	return 0;
}

static yk0b_error_t yk0b_procedure_get_priority(yk0b_task_link_t link) {
	yk0b_task_t t;
	t = yk0b_get_priority_read_task(link);
	if (t >= YK0B_SETUP_TASK_SIZE) {
		return 1;
	}
	yk0b_get_priority_write_priority(link, yk0b_task_read_priority(yk0b_task_to_task_link(t)));
	return 0;
}

static yk0b_error_t yk0b_procedure_listen(yk0b_task_link_t link) {
	yk0b_flag_t f;
	yk0b_port_t p;
	p = yk0b_listen_read_port(link);
	if (p >= YK0B_SETUP_PORT_SIZE) {
		return 1;
	}
	if ((yk0b_port_read_flag(p) & YK0B_FLAG_KERNEL_RECEIVER) != 0) {
		return 1;
	}
	f = yk0b_listen_read_flag(link);
	if (((yk0b_port_read_flag(p) | YK0B_FLAG_KERNEL_RECEIVER) & f) != 0) {
		yk0b_listen_write_flag(link, yk0b_port_read_flag(p) | YK0B_FLAG_KERNEL_RECEIVER);
		yk0b_port_write_flag(p, yk0b_port_read_flag(p) & (~f | YK0B_FLAG_KERNEL_SENDER));
	} else {
		yk0b_schedule_delete(link);
		yk0b_task_write_port(link, p);
		yk0b_task_write_state(link, YK0B_STATE_SEND);
		yk0b_port_write_flag(p, yk0b_port_read_flag(p) | YK0B_FLAG_KERNEL_RECEIVER);
		yk0b_port_write_receiver(p, link);
		yk0b_schedule_update();
	}
	return 0;
}

static yk0b_error_t yk0b_procedure_reply(yk0b_task_link_t link) {
	yk0b_task_t t;
	t = yk0b_reply_read_task(link);
	if (t >= YK0B_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0b_task_read_state(yk0b_task_to_task_link(t)) != YK0B_STATE_REPLY) {
		return 1;
	}
	if (yk0b_task_read_replyer(yk0b_task_to_task_link(t)) != link) {
		return 1;
	}
	yk0b_task_write_state(yk0b_task_to_task_link(t), YK0B_STATE_READY);
	yk0b_schedule_append(yk0b_task_to_task_link(t));
	yk0b_schedule_update();
	return 0;
}

static yk0b_error_t yk0b_procedure_set_context(yk0b_task_link_t link) {
	yk0b_context_type_t ct;
	yk0b_context_value_t cv;
	yk0b_task_t t;
	t = yk0b_set_context_read_task(link);
	if (t >= YK0B_SETUP_TASK_SIZE) {
		return 1;
	}
	if (yk0b_task_read_state(yk0b_task_to_task_link(t)) != YK0B_STATE_REPLY) {
		return 1;
	}
	if (yk0b_task_read_replyer(yk0b_task_to_task_link(t)) != link) {
		return 1;
	}
	ct = yk0b_set_context_read_type(link);
	if (ct >= YK0B_CONTEXT_SIZE) {
		return 1;
	}
	cv = yk0b_set_context_read_value(link);
	yk0b_set_context_write_value(link, yk0b_task_read_context(yk0b_task_to_task_link(t), ct));
	yk0b_task_write_context(yk0b_task_to_task_link(t), ct, cv);
	return 0;
}

static yk0b_error_t yk0b_procedure_set_priority(yk0b_task_link_t link) {
	yk0b_priority_t p;
	yk0b_task_t t;
	t = yk0b_set_priority_read_task(link);
	if (t >= YK0B_SETUP_TASK_SIZE) {
		return 1;
	}
	p = yk0b_set_priority_read_priority(link);
	if (p >= YK0B_SETUP_PRIORITY_SIZE) {
		return 1;
	}
	yk0b_set_priority_write_priority(link, yk0b_task_read_priority(yk0b_task_to_task_link(t)));
	if (yk0b_task_read_state(yk0b_task_to_task_link(t)) == YK0B_STATE_READY) {
		yk0b_schedule_delete(yk0b_task_to_task_link(t));
	}
	yk0b_task_write_priority(yk0b_task_to_task_link(t), p);
	if (yk0b_task_read_state(yk0b_task_to_task_link(t)) == YK0B_STATE_READY) {
		yk0b_schedule_append(yk0b_task_to_task_link(t));
		yk0b_schedule_update();
	}
	return 0;
}

/* external function */

void yk0b_core_initialize(void) {
	yk0b_task_t t;
	yk0b_port_initialize();
	yk0b_schedule_initialize();
	for (t = 0; t < YK0B_SETUP_TASK_SIZE; ++t) {
		yk0b_task_write_priority(yk0b_task_to_task_link(t), 0);
		if (t == 0) {
			yk0b_task_write_state(yk0b_task_to_task_link(t), YK0B_STATE_READY);
			yk0b_schedule_append(yk0b_task_to_task_link(t));
		} else {
			yk0b_task_write_replyer(yk0b_task_to_task_link(t), yk0b_task_to_task_link(0));
			yk0b_task_write_state(yk0b_task_to_task_link(t), YK0B_STATE_REPLY);
		}
	}
	yk0b_schedule_update();
}

void cin_task(void);

void yk0b_core_exception_interrupt_procedure(yk0b_uint32_t cause) {
	yk0b_task_link_t l;
	switch (cause & YK0B_MIPS_CAUSE_EXCCODE_MASK) {
	case YK0B_MIPS_CAUSE_EXCCODE_INT:
		cin_task();
		break;
	case YK0B_MIPS_CAUSE_EXCCODE_SYS:
		l = yk0b_schedule_link;
		yk0b_task_write_context(l, YK0B_CONTEXT_EPC, yk0b_task_read_context(l, YK0B_CONTEXT_EPC) + 4); /* ignore YK0B_MIPS_CAUSE_BD */
		switch (yk0b_task_read_context(l, YK0B_CONTEXT_ARGUMENT0)) {
		case YK0B_PROCEDURE_ACCEPT:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_accept(l));
			break;
		case YK0B_PROCEDURE_BIT_FLAG:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_bit_flag(l));
			break;
		case YK0B_PROCEDURE_CALL:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_call(l));
			cin_task();
			break;
		case YK0B_PROCEDURE_CONNECT:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_connect(l));
			break;
		case YK0B_PROCEDURE_FORWARD:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_forward(l));
			cin_task();
			break;
		case YK0B_PROCEDURE_GET_CONTEXT:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_get_context(l));
			break;
		case YK0B_PROCEDURE_GET_PRIORITY:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_get_priority(l));
			break;
		case YK0B_PROCEDURE_LISTEN:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_listen(l));
			break;
		case YK0B_PROCEDURE_REPLY:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_reply(l));
			break;
		case YK0B_PROCEDURE_SET_CONTEXT:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_set_context(l));
			break;
		case YK0B_PROCEDURE_SET_PRIORITY:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, yk0b_procedure_set_priority(l));
			break;
		default:
			yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT0, 1);
			break;
		}
		break;
	default:
		yk0b_task_write_context(yk0b_schedule_link, YK0B_CONTEXT_CAUSE, cause);
		yk0b_exception_forward(YK0B_SETUP_EXCEPTION_PORT, yk0b_schedule_link);
		break;
	}
}

#include "device.h"
#include "yanamips.h"

typedef signed char y_int8_t;
typedef signed short y_int16_t;
typedef signed int y_int32_t;
typedef unsigned char y_uint1_t;
typedef unsigned char y_uint5_t;
typedef unsigned char y_uint7_t;
typedef unsigned char y_uint8_t;
typedef unsigned short y_uint16_t;
typedef unsigned int y_uint26_t;
typedef unsigned int y_uint32_t;

union {
	y_uint32_t w;
	y_uint8_t b[4];
} m_main[4096] = {
#include "m_main.h"
};

y_uint8_t cin;

y_uint32_t r_gpr[32];

y_uint32_t r_badvaddr; /* 8 */
y_uint32_t r_status; /* 12 */
y_uint32_t r_cause; /* 13 */
y_uint32_t r_epc; /* 14 */

y_uint32_t r_pc;
y_uint1_t r_bd;
y_uint32_t r_next;

void cin_task(void) {
	if (cin != 0 && (yk0b_port_read_flag(YK0B_SETUP_INTERRUPT_PORT) & YK0B_FLAG_KERNEL_SENDER) != 0) {
		yk0b_task_link_t l;
		l = yk0b_port_remove(YK0B_SETUP_INTERRUPT_PORT);
		yk0b_task_write_context(l, YK0B_CONTEXT_ARGUMENT1, cin);
		cin = 0;
		r_cause &= ~(DEVICE_INTERRUPT_CONS << YANAMIPS_CAUSE_IP_SHIFT);
		yk0b_task_write_state(l, YK0B_STATE_READY);
		yk0b_schedule_append(l);
		yk0b_schedule_update();
	}
}

y_uint32_t handle(y_uint7_t exccode) {
	r_cause = r_cause & ~YANAMIPS_CAUSE_EXCCODE_MASK | exccode;
	if ((r_status & YANAMIPS_STATUS_EXL) == 0) {
		r_status |= YANAMIPS_STATUS_EXL;
		if (r_bd) {
			r_cause |= YANAMIPS_CAUSE_BD;
			r_epc = r_pc - 4;
		} else {
			r_cause &= ~YANAMIPS_CAUSE_BD;
			r_epc = r_pc;
		}
	}
	if (yk0b_schedule_link != QUQUE_NULL) {
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_EPC] = r_epc;
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR1] = r_gpr[1];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR2] = r_gpr[2];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR3] = r_gpr[3];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR4] = r_gpr[4];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR5] = r_gpr[5];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR6] = r_gpr[6];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR7] = r_gpr[7];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR8] = r_gpr[8];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR9] = r_gpr[9];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR10] = r_gpr[10];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR11] = r_gpr[11];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR12] = r_gpr[12];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR13] = r_gpr[13];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR14] = r_gpr[14];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR15] = r_gpr[15];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR16] = r_gpr[16];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR17] = r_gpr[17];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR18] = r_gpr[18];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR19] = r_gpr[19];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR20] = r_gpr[20];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR21] = r_gpr[21];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR22] = r_gpr[22];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR23] = r_gpr[23];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR24] = r_gpr[24];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR25] = r_gpr[25];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR28] = r_gpr[28];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR29] = r_gpr[29];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR30] = r_gpr[30];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR31] = r_gpr[31];
		yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_STATUS] = r_status;
	}
	yk0b_core_exception_interrupt_procedure(r_cause);
	if (yk0b_schedule_link == QUQUE_NULL) {
		r_status = r_status & ~(YK0B_MIPS_STATUS_KSU_MASK | YK0B_MIPS_STATUS_EXL) | YK0B_MIPS_STATUS_IE;
		return ~0;
	} else {
		r_epc = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_EPC];
		r_gpr[1] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR1];
		r_gpr[2] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR2];
		r_gpr[3] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR3];
		r_gpr[4] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR4];
		r_gpr[5] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR5];
		r_gpr[6] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR6];
		r_gpr[7] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR7];
		r_gpr[8] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR8];
		r_gpr[9] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR9];
		r_gpr[10] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR10];
		r_gpr[11] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR11];
		r_gpr[12] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR12];
		r_gpr[13] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR13];
		r_gpr[14] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR14];
		r_gpr[15] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR15];
		r_gpr[16] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR16];
		r_gpr[17] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR17];
		r_gpr[18] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR18];
		r_gpr[19] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR19];
		r_gpr[20] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR20];
		r_gpr[21] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR21];
		r_gpr[22] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR22];
		r_gpr[23] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR23];
		r_gpr[24] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR24];
		r_gpr[25] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR25];
		r_gpr[28] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR28];
		r_gpr[29] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR29];
		r_gpr[30] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR30];
		r_gpr[31] = yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_GPR31];
		r_status = r_status & (YK0B_MIPS_STATUS_IM_MASK | YK0B_MIPS_STATUS_ERL) | yk0b_task_control[yk0b_schedule_link].context[YK0B_CONTEXT_STATUS] & ~(YK0B_MIPS_STATUS_IM_MASK | YK0B_MIPS_STATUS_ERL | YK0B_MIPS_STATUS_EXL);
		return r_epc;
	}
}

y_uint8_t hlscore(y_uint8_t reset, y_uint8_t c) {
	y_uint8_t cout = 0;
	y_uint32_t code;
	y_uint26_t c2i;
	y_uint5_t c4s;
	y_uint5_t c4t;
	y_uint16_t c4i;
	y_uint5_t c6s;
	y_uint5_t c6t;
	y_uint5_t c6d;
	y_uint5_t c6h;
	y_uint1_t bd;
	y_uint32_t next;
	y_uint32_t i;
	y_uint32_t j;
	if (reset != 0) {
		r_gpr[0] = 0;
		r_cause = 0;
		r_status = 0;
		r_pc = 0;
		r_bd = 0;
		r_next = r_pc + 4;
		yk0b_core_initialize();
		return 0;
	}
	if (c != 0) {
		cin = c;
		r_cause |= DEVICE_INTERRUPT_CONS << YANAMIPS_CAUSE_IP_SHIFT;
	}
	if ((r_cause >> YANAMIPS_CAUSE_IP_SHIFT & (r_status & YANAMIPS_STATUS_IM_MASK) >> YANAMIPS_STATUS_IM_SHIFT) != 0 && (r_status & (YANAMIPS_STATUS_EXL | YANAMIPS_STATUS_IE)) == YANAMIPS_STATUS_IE) {
		r_pc = handle(YANAMIPS_CAUSE_EXCCODE_INT);
		r_bd = 0;
		r_next = r_pc + 4;
		return 0;
	}
	if (r_pc == ~0) {
		return 0;
	}
	bd = 0;
	next = r_next + 4;
	/* ToDo: check r_pc */
	code = m_main[r_pc >> 2].w;
	c2i = code & 0x03ffffff;
	c4s = code >> 21 & 0x1f;
	c4t = code >> 16 & 0x1f;
	c4i = code & 0xffff;
	c6s = code >> 21 & 0x1f;
	c6t = code >> 16 & 0x1f;
	c6d = code >> 11 & 0x1f;
	c6h = code >> 6 & 0x1f;
	switch (code >> 26) {
	case 0x00:
		switch (code & 0x3f) {
		case 0x00: /* sll */
			/* ToDo: check c6h */
			i = r_gpr[c6t] << c6h;
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x02: /* srl */
			/* ToDo: check c6h */
			i = r_gpr[c6t] >> c6h;
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x03: /* sra */
			/* ToDo: check c6h */
			i = (y_int32_t)r_gpr[c6t] >> c6h;
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x24: /* and */
			i = r_gpr[c6s] & r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x25: /* or */
			i = r_gpr[c6s] | r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x26: /* xor */
			i = r_gpr[c6s] ^ r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x08: /* jr */
			/* ToDo: check r_bd */
			bd = 1;
			next = r_gpr[c6s];
			break;
		case 0x09: /* jalr */
			/* ToDo: check r_bd */
			i = r_next + 4;
			j = r_gpr[c6s];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			bd = 1;
			next = j;
			break;
		case 0x0c: /* syscall */
			r_next = handle(YANAMIPS_CAUSE_EXCCODE_SYS);
			next = r_next + 4;
			break;
		case 0x21: /* addu */
			i = r_gpr[c6s] + r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x23: /* subu */
			i = r_gpr[c6s] - r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x2a: /* slt */
			i = (y_int32_t)r_gpr[c6s] < (y_int32_t)r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		case 0x2b: /* sltu */
			i = r_gpr[c6s] < r_gpr[c6t];
			if (c6d != 0) {
				r_gpr[c6d] = i;
			}
			break;
		default:
			/* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
			break;
		}
		break;
	case 0x02: /* j */
		/* ToDo: check r_bd */
		bd = 1;
		next = r_next & 0xf0000000 | c2i << 2;
		break;
	case 0x03: /* jal */
		/* ToDo: check r_bd */
		r_gpr[31] = r_next + 4;
		bd = 1;
		next = r_next & 0xf0000000 | c2i << 2;
		break;
	case 0x04: /* beq */
		/* ToDo: check r_bd */
		bd = 1;
		if (r_gpr[c4s] == r_gpr[c4t]) {
			next = r_next + ((y_int16_t)c4i << 2);
		}
		break;
	case 0x05: /* bne */
		/* ToDo: check r_bd */
		bd = 1;
		if (r_gpr[c4s] != r_gpr[c4t]) {
			next = r_next + ((y_int16_t)c4i << 2);
		}
		break;
	case 0x09: /* addiu */
		i = r_gpr[c4s] + (y_int16_t)c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0a: /* slti */
		i = (y_int32_t)r_gpr[c4s] < (y_int16_t)c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0b: /* sltiu */
		i = r_gpr[c4s] < (y_int16_t)c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0c: /* andi */
		i = r_gpr[c4s] & c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0d: /* ori */
		i = r_gpr[c4s] | c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0e: /* xori */
		i = r_gpr[c4s] ^ c4i;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x0f: /* lui */
		i = c4i << 16;
		if (c4t != 0) {
			r_gpr[c4t] = i;
		}
		break;
	case 0x10:
		switch (c6s) {
		case 0x00: /* mfc0 */
			switch (c6d) {
			case 8:
				i = r_badvaddr;
				break;
			case 12:
				i = r_status;
				break;
			case 13:
				i = r_cause;
				break;
			case 14:
				i = r_epc;
				break;
			default:
				/* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
				break;
			}
			if (c6t != 0) {
				r_gpr[c6t] = i;
			}
			break;
		case 0x04: /* mtc0 */
			switch (c6d) {
			case 8:
				break;
			case 12:
				r_status = r_status & ~(YANAMIPS_STATUS_IM_MASK | YANAMIPS_STATUS_EXL | YANAMIPS_STATUS_IE) | r_gpr[c6t] & (YANAMIPS_STATUS_IM_MASK | YANAMIPS_STATUS_EXL | YANAMIPS_STATUS_IE);
				break;
			case 13:
				r_cause = r_cause & ~(YANAMIPS_CAUSE_IP1 | YANAMIPS_CAUSE_IP0) | r_gpr[c6t] & (YANAMIPS_CAUSE_IP1 | YANAMIPS_CAUSE_IP0);
				break;
			case 14:
				r_epc = r_gpr[c6t];
				break;
			default:
				/* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
				break;
			}
			break;
		case 0x10: /* eret */
			/* ToDo: check r_bd */
			r_status &= ~YANAMIPS_STATUS_EXL;
			r_next = r_epc;
			next = r_next + 4;
			break;
		default:
			/* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
			break;
		}
		break;
	case 0x20: /* lb */
		switch (r_gpr[c4s] + (y_int16_t)c4i) {
		case DEVICE_CONS_PUTGETCHAR:
			i = cin;
			if (c4t != 0) {
				r_gpr[c4t] = i;
			}
			cin = 0;
			r_cause &= ~(DEVICE_INTERRUPT_CONS << YANAMIPS_CAUSE_IP_SHIFT);
			break;
		case DEVICE_CONS_HALT:
			/* ToDo: halt */
			break;
		default:
#if 0
			i = m_main[r_gpr[c4s] + (y_int16_t)c4i >> 2].b[r_gpr[c4s] + (y_int16_t)c4i & 3];
#else
			i = m_main[r_gpr[c4s] + (y_int16_t)c4i >> 2].b[~(r_gpr[c4s] + (y_int16_t)c4i) & 3];
#endif
			if (c4t != 0) {
				r_gpr[c4t] = i;
			}
			break;
		}
		break;
	case 0x23: /* lw */
		if ((r_gpr[c4s] + (y_int16_t)c4i & 3) != 0) {
			r_badvaddr = r_gpr[c4s] + (y_int16_t)c4i;
			r_next = handle(YANAMIPS_CAUSE_EXCCODE_ADEL);
			next = r_next + 4;
		} else {
			i = m_main[(r_gpr[c4s] + (y_int16_t)c4i) >> 2].w;
			if (c4t != 0) {
				r_gpr[c4t] = i;
			}
		}
		break;
	case 0x28: /* sb */
		switch (r_gpr[c4s] + (y_int16_t)c4i) {
		case DEVICE_CONS_PUTGETCHAR:
			cout = r_gpr[c4t];
			break;
		case DEVICE_CONS_HALT:
			/* ToDo: halt */
			break;
		default:
#if 0
			m_main[r_gpr[c4s] + (y_int16_t)c4i >> 2].b[r_gpr[c4s] + (y_int16_t)c4i & 3] = r_gpr[c4t];
#else
			m_main[r_gpr[c4s] + (y_int16_t)c4i >> 2].b[~(r_gpr[c4s] + (y_int16_t)c4i) & 3] = r_gpr[c4t];
#endif
			break;
		}
		break;
	case 0x2b: /* sw */
		if ((r_gpr[c4s] + (y_int16_t)c4i & 3) != 0) {
			r_badvaddr = r_gpr[c4s] + (y_int16_t)c4i;
			r_next = handle(YANAMIPS_CAUSE_EXCCODE_ADES);
			next = r_next + 4;
		} else {
			m_main[(r_gpr[c4s] + (y_int16_t)c4i) >> 2].w = r_gpr[c4t];
		}
		break;
	default:
		 /* ToDo: YANAMIPS_CAUSE_EXCCODE_RI */
		break;
	}
	r_pc = r_next;
	r_bd = bd;
	r_next = next;
	return cout;
}
